%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     This code performs the SMC sampler algorithm
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   Matlab Codes associated to the paper
%       ``A Bayesian Perspective on Multiple Source
%        Localization in Wireless Sensor Network'' 
% by T.L.T Nguyen, F. Septier, H. Rajaona, G.W. Peters, I. Nevat,
%       Y. Delignon
%  
%       Date: Sep. 2015
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [SMCSampler]=SMCSamplerAlgo(SMCSampler,Target,CurrentNbMCMCMoves,CurrentNbBlockMWG,Sensors,PercentageCESS);




% Function Used to evaluate the log Inverse Gamma (prior of the Power)
LogPrior_LogValue=@(x,Shape0,Scale0) -Scale0./(exp(x))-Shape0*x+Shape0*log(Scale0);


% ---- Definition of the Parameters for Prior
Prior.Beta.Mean=Target.PriorMean;
Prior.Beta.Cov=Target.PriorCov;
Prior.Beta.Dimension=Target.Dim;

Shape0=Target.PowerShape;
Scale0=Target.PowerScale;

% Initialization of internal SMC parameters
SMCSampler.NbIter=size(SMCSampler.Temperature,2);
SMCSampler.NbMCMCMoves=CurrentNbMCMCMoves;
SMCSampler.NormalizedWeights=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);
SMCSampler.LogWeights=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);
SMCSampler.LogWeightsSaved=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);
SMCSampler.LogLike=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);
SMCSampler.SavedLogLike=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);

% Defition of the Metropolis Hastings Within Gibbs Step

SMCSampler.NbBlockMWG=CurrentNbBlockMWG;% Number of MCMC Moves
CovarianceRWLogPower=3; % Variance Used in the Random Walk of the log Power
CovarianceRWBeta=10; % Variance Used in the Random Walk of the position


SMCSampler.LogPriorBeta=zeros(1,SMCSampler.NbPart,SMCSampler.NbIter);
SMCSampler.AcceptedMove.Beta=0;

SMCSampler.NbTotalMCMCMoves=0;
SMCSampler.NbResampling=0;


SMCSampler.AcceptedMove.Beta=zeros(SMCSampler.NbBlockMWG,SMCSampler.NbIter);
SMCSampler.AcceptedMove.LogPower=zeros(Target.Nb,SMCSampler.NbIter);
SMCSampler.SizeBlockMWG=Prior.Beta.Dimension/SMCSampler.NbBlockMWG;
SMCSampler.ResamplingDoneatThisIteration=zeros(1,SMCSampler.NbIter);


SMCSampler.ResamplingDoneatThisIteration(1)=1; % Equally Weighted At the first Iteration !
SMCSampler.NumIter=0;

LogRatioNormalizing=zeros(1,SMCSampler.NbIter);

SMCSampler.Temperature=zeros(1,SMCSampler.NbIter);









CurrentTemperature=0;


 while  CurrentTemperature<1 % Iterate until the temperature is below 1
    
    SMCSampler.NumIter=SMCSampler.NumIter+1;
    t=SMCSampler.NumIter;

    
    if t>1
        
        % ---- Adaptive procedure for the temperature based on the CESS !!
        
        fun = @(x) Evaluating_CESS(x,SMCSampler,PercentageCESS,t);
        x0 = 1e-3; % initial point
        [OptNewIncrements,FVAL] = fzero(fun,x0);
        
        OptNewIncrements=abs(OptNewIncrements);
        
        
        SMCSampler.Temperature(t)=SMCSampler.Temperature(t-1)+OptNewIncrements;
        if SMCSampler.Temperature(t)>1
            SMCSampler.Temperature(t)=1;
        end
        %figure(15);plot(SMCSampler.Temperature(1:t))
        CurrentTemperature=SMCSampler.Temperature(t);
        
              
 
        % Computation of the importance weights (in log)
        SMCSampler.LogWeights(1,:,t)=log(SMCSampler.NormalizedWeights(1,:,t-1))+...
                   (SMCSampler.Temperature(t)-SMCSampler.Temperature(t-1))*SMCSampler.LogLike(1,:,t-1); 
            
        
        
        SMCSampler.LogWeightsSaved(1,:,t)=SMCSampler.LogWeights(1,:,t);
        
        % Normalization of the importance weights
        
        MaximumValue=max(SMCSampler.LogWeights(1,:,t));
        SMCSampler.NormalizedWeights(1,:,t)=exp(SMCSampler.LogWeights(1,:,t)-MaximumValue)/sum(exp(SMCSampler.LogWeights(1,:,t)-MaximumValue));


        % Computation of the Effective Sample Size (ESS)
        Neff=1/sum(abs(SMCSampler.NormalizedWeights(1,:,t)).^2);
        SMCSampler.EffectiveSampleSize(t)=Neff;
        % RESAMPLING STEP
        if (Neff<SMCSampler.NbPart/2) || SMCSampler.ResamplingEachIteration==1% DO RESAMPLING         
           % disp('RESAMPLING')
            SMCSampler.NbResampling=SMCSampler.NbResampling+1;
            outIndex = residualR(1:SMCSampler.NbPart,SMCSampler.NormalizedWeights(1,:,t)');
            
            
            SMCSampler.Iter(t-1).Particles=SMCSampler.Iter(t-1).Particles(outIndex);

            SMCSampler.LogPriorBeta(1,:,t-1)=SMCSampler.LogPriorBeta(1,outIndex,t-1);
            SMCSampler.LogPriorLogPower(1,:,t-1)=SMCSampler.LogPriorLogPower(1,outIndex,t-1);           
            SMCSampler.LogLike(1,:,t-1)=SMCSampler.LogLike(1,outIndex,t-1);
            
            
            SMCSampler.LogWeights(1,:,t)=log(1/SMCSampler.NbPart);
            SMCSampler.NormalizedWeights(1,:,t)=1/SMCSampler.NbPart;   
            
            SMCSampler.ResamplingDoneatThisIteration(1,t)=1;
        end        
        
    
        
    end
    
    
    for j=1:SMCSampler.NbPart % Iterations over the particles
        
        % INITIALIZATION
        if t==1
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % --- DRAW FROM THE PRIOR ---%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            

            
            
            % Draw Beta From Prior
            SMCSampler.Iter(t).Particles(j).Beta=Prior.Beta.Mean+sqrtm(Prior.Beta.Cov)*randn(Prior.Beta.Dimension,1);
            SMCSampler.Iter(t).Particles(j).LogPower=log(1./gamrnd(Shape0,1/Scale0,Target.Nb,1));

            
            % Computation of the prior for the position (in log)
            SMCSampler.LogPriorBeta(1,j,t)=ComputationLogNormalPDF(SMCSampler.Iter(t).Particles(j).Beta,Prior.Beta.Mean,Prior.Beta.Cov);
            
            % Computation of the prior for the power (in log)
            SMCSampler.LogPriorLogPower(1,j,t)=sum(LogPrior_LogValue(SMCSampler.Iter(t).Particles(j).LogPower,Shape0,Scale0));

            % Computation of the likelihood (in log)
            SMCSampler.LogLike(1,j,t)=ComputationLogLikelihood(Sensors,SMCSampler.Iter(t).Particles(j).Beta,Target,exp(SMCSampler.Iter(t).Particles(j).LogPower));
            
            SMCSampler.LogWeights(1,j,t)=0;
            SMCSampler.NormalizedWeights(1,j,t)=1/SMCSampler.NbPart;
            
            SMCSampler.SavedLogLike(1,j,t)=SMCSampler.LogLike(1,j,t);   % Just to keep it !!
        else
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % ---     MCMC Moves      ---%
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%            
            
            CurrentBeta=SMCSampler.Iter(t-1).Particles(j).Beta;
            CurrentLogPower=SMCSampler.Iter(t-1).Particles(j).LogPower;

            CurrentLogPriorBeta=SMCSampler.LogPriorBeta(1,j,t-1);
            CurrentLogPriorLogPower=SMCSampler.LogPriorLogPower(1,j,t-1);
            CurrentLogLike=SMCSampler.LogLike(1,j,t-1);
            

            for n=1:SMCSampler.NbMCMCMoves % Loop over the number of MCMC Moves

                
                Permutation=randperm(SMCSampler.NbBlockMWG); % Random Permutation of the order of the source to sample
                
                SMCSampler.NbTotalMCMCMoves=SMCSampler.NbTotalMCMCMoves+1; % To Count AcceptanceRate
                
                % ---- MH within Gibbs for the Power
                
                for BlockIndex=1:Target.Nb
                    %Permutation(BlockIndex)
                    Index_Element_Block=BlockIndex;%(Permutation(BlockIndex)-1)*SMCSampler.SizeBlockMWG+1:Permutation(BlockIndex)*SMCSampler.SizeBlockMWG;
                    
                    % --- Adaptation of the Covariance of the Random Walk Proposal Based on previous population
                    %CovarianceRW=diag(diag(SMCSampler.Estimate.Cov(Index_Element_Block,Index_Element_Block)));
                    CovarianceRW=diag(diag(SMCSampler.Estimate.CovLogPower(Index_Element_Block,Index_Element_Block)));
                    CovarianceRW=CovarianceRWLogPower*eye(length(Index_Element_Block));

                    if t>2
                        % Readjust the Covariance with pair Threshold
                        RateAcceptance=SMCSampler.AcceptedMove.LogPower(Permutation(BlockIndex),t-1)/(SMCSampler.NbMCMCMoves*SMCSampler.NbPart);
                        if RateAcceptance>0.7
                            CovarianceRW=CovarianceRW*5;
                        elseif RateAcceptance<0.15
                            CovarianceRW=CovarianceRW*(1/5);
                        end
                    end
                    
                    
                    
                    CandidateLogPower=CurrentLogPower;

                    CandidateLogPower(Index_Element_Block,1)=CandidateLogPower(Index_Element_Block,1)+sqrtm(CovarianceRW)*randn;
                    
                    CandidateLogPriorLogPower=sum(LogPrior_LogValue(CandidateLogPower,Shape0,Scale0));

                    
                    CandidateLogLike=ComputationLogLikelihood(Sensors,CurrentBeta,Target,exp(CandidateLogPower));
                    AccRatio=min([1 exp(SMCSampler.Temperature(t)*(CandidateLogLike-CurrentLogLike)+CandidateLogPriorLogPower-CurrentLogPriorLogPower)]);

                    % Acceptance
                    if rand<AccRatio
                       CurrentLogPower=CandidateLogPower;
                       CurrentLogPriorLogPower=CandidateLogPriorLogPower;
                       CurrentLogLike=CandidateLogLike;

                       % -- To Analyze Acceptance Rate for each Block
                       SMCSampler.AcceptedMove.LogPower(Index_Element_Block,t)=SMCSampler.AcceptedMove.LogPower(Index_Element_Block,t)+1;
                    end                       
                
                end                               
            
                % ---- MH within Gibbs for the position
                
                for BlockIndex=1:SMCSampler.NbBlockMWG
                    %Permutation(BlockIndex)
                    Index_Element_Block=(Permutation(BlockIndex)-1)*SMCSampler.SizeBlockMWG+1:Permutation(BlockIndex)*SMCSampler.SizeBlockMWG;
                    
                    % --- Adaptation of the Covariance of the Random Walk Proposal Based on previous population
                    CovarianceRW=diag(diag(SMCSampler.Estimate.Cov(Index_Element_Block,Index_Element_Block)));
                    CovarianceRW=CovarianceRWBeta*eye(length(Index_Element_Block));

                    if t>2
                        % Readjust the Covariance with pair Threshold
                        RateAcceptance=SMCSampler.AcceptedMove.Beta(Permutation(BlockIndex),t-1)/(SMCSampler.NbMCMCMoves*SMCSampler.NbPart);
                        if RateAcceptance>0.7
                            CovarianceRW=CovarianceRW*5;
                        elseif RateAcceptance<0.15
                            CovarianceRW=CovarianceRW*(1/5);
                        end
                    end
                    
                    
                    CandidateBeta=CurrentBeta;

                    CandidateBeta(Index_Element_Block,1)=CandidateBeta(Index_Element_Block,1)+sqrtm(CovarianceRW)*randn(SMCSampler.SizeBlockMWG,1);
                    
                    
                    CandidateLogPriorBeta=ComputationLogNormalPDF(CandidateBeta,Prior.Beta.Mean,Prior.Beta.Cov);

                    CandidateLogLike=ComputationLogLikelihood(Sensors,CandidateBeta,Target,exp(CurrentLogPower));
                    AccRatio=min([1 exp(SMCSampler.Temperature(t)*(CandidateLogLike-CurrentLogLike)+CandidateLogPriorBeta-CurrentLogPriorBeta)]);
                
                    % Acceptance
                    if rand<AccRatio
                       CurrentBeta=CandidateBeta;
                       CurrentLogPriorBeta=CandidateLogPriorBeta;
                       CurrentLogLike=CandidateLogLike;

                       % -- To Analyze Acceptance Rate for each Block
                       SMCSampler.AcceptedMove.Beta(Permutation(BlockIndex),t)=SMCSampler.AcceptedMove.Beta(Permutation(BlockIndex),t)+1;
                    end                       
                
                end
                      

                
            end
            
            % Save the last output of the MCMC moves to be the current
            % particl values

            SMCSampler.Iter(t).Particles(j).Beta=CurrentBeta;
            SMCSampler.Iter(t).Particles(j).LogPower=CurrentLogPower;
            
            SMCSampler.LogPriorBeta(1,j,t)=CurrentLogPriorBeta;
            SMCSampler.LogPriorLogPower(1,j,t)=CurrentLogPriorLogPower;

            SMCSampler.LogLike(1,j,t)=CurrentLogLike;            
            
            SMCSampler.SavedLogLike(1,j,t)=SMCSampler.LogLike(1,j,t);   % Just to keep it !!
            
            
        end
        
    end
    

    % Computation of the ratio of normalizing contant (in log)
    LogRatioNormalizing(1,t)=log(sum(exp(SMCSampler.LogWeightsSaved(1,:,t))));
    
    SMCSampler.Iter(t).SavedParticles=SMCSampler.Iter(t).Particles; % To be used for computing Estimation based on all the particles (Recycling)
    
    if 1 %Computation of the mean and covariance of the current target distributions
        Weights=repmat(SMCSampler.NormalizedWeights(1,:,t),Prior.Beta.Dimension,1);
        WeightsPower=repmat(SMCSampler.NormalizedWeights(1,:,t),Target.Nb,1);        
        BetaParticlesValues=[SMCSampler.Iter(t).Particles(:).Beta];
        
        LogPowerParticlesValues=[SMCSampler.Iter(t).Particles(:).LogPower];
        
        SMCSampler.Estimate.Mean=sum(Weights.*BetaParticlesValues,2);
        
        BetaParticlesCentered=BetaParticlesValues-repmat(SMCSampler.Estimate.Mean,1,SMCSampler.NbPart);
        SMCSampler.Estimate.Cov=Weights.*BetaParticlesCentered*BetaParticlesCentered';
        
        SMCSampler.Estimate.MeanLogPower=sum(WeightsPower.*LogPowerParticlesValues,2);
        
        LogPowerParticlesCentered=LogPowerParticlesValues-repmat(SMCSampler.Estimate.MeanLogPower,1,SMCSampler.NbPart);
        SMCSampler.Estimate.CovLogPower=WeightsPower.*LogPowerParticlesCentered*LogPowerParticlesCentered';        
        

    end
   
    
end
% Compute the final estimate of the Evidence (Marginal Likelihood)
LogBayesEvidence=sum(LogRatioNormalizing,2)-log(SMCSampler.NbPart);
SMCSampler.LogBayesEvidence=LogBayesEvidence;